#include "LoudnessCalcThread.h"

#include <chrono>

#include "DatabaseBeatmap.h"
#include "Engine.h"
#include "Sound.h"
#include "SoundEngine.h"

struct LoudnessCalcThread {
    std::thread thr;
    std::atomic<bool> dead = true;
    std::vector<BeatmapDifficulty*> maps;

    std::atomic<u32> nb_computed = 0;
    std::atomic<u32> nb_total = 0;

   public:
    LoudnessCalcThread(std::vector<BeatmapDifficulty*> maps_to_calc) {
        dead = false;
        maps = maps_to_calc;
        nb_total = maps.size() + 1;
        thr = std::thread(&LoudnessCalcThread::run, this);
    }

    ~LoudnessCalcThread() {
        dead = true;
        thr.join();
    }

   private:
    void run() {
        std::string last_song = "";
        f32 last_loudness = 0.f;
        f32 buf[44100];

        BASS_SetDevice(0);
        BASS_SetConfig(BASS_CONFIG_UPDATETHREADS, 0);

        for(auto diff2 : maps) {
            while(osu->should_pause_background_threads.load() && !dead.load()) {
                std::this_thread::sleep_for(std::chrono::milliseconds(100));
            }

            if(dead.load()) return;
            if(diff2->loudness.load() != 0.f) continue;

            auto song = diff2->getFullSoundFilePath();
            if(song == last_song) {
                diff2->loudness = last_loudness;
                nb_computed++;
                continue;
            }

            auto decoder = BASS_StreamCreateFile(false, song.c_str(), 0, 0, BASS_STREAM_DECODE | BASS_SAMPLE_MONO);
            if(!decoder) {
                debugLog("BASS_StreamCreateFile() returned error %d on file %s\n", BASS_ErrorGetCode(), song.c_str());
                nb_computed++;
                continue;
            }

            auto loudness = BASS_Loudness_Start(decoder, BASS_LOUDNESS_INTEGRATED, 0);
            if(!loudness) {
                debugLog("BASS_Loudness_Start() returned error %d\n", BASS_ErrorGetCode());
                BASS_ChannelFree(decoder);
                nb_computed++;
                continue;
            }

            // Did you know?
            // If we do while(BASS_ChannelGetData(decoder, buf, sizeof(buf) >= 0), we get an infinite loop!
            // Thanks, Microsoft!
            int c;
            do {
                c = BASS_ChannelGetData(decoder, buf, sizeof(buf));
            } while(c >= 0);

            BASS_ChannelFree(decoder);

            f32 integrated_loudness = 0.f;
            BASS_Loudness_GetLevel(loudness, BASS_LOUDNESS_INTEGRATED, &integrated_loudness);
            BASS_Loudness_Stop(loudness);
            if(integrated_loudness == -HUGE_VAL) {
                debugLog("No loudness information available for '%s' (silent song?)\n", song.c_str());
            } else {
                diff2->loudness = integrated_loudness;
                diff2->update_overrides();
                last_loudness = integrated_loudness;
                last_song = song;
            }

            nb_computed++;
        }

        nb_computed++;
    }
};

static std::vector<LoudnessCalcThread*> threads;

u32 loct_computed() {
    u32 x = 0;
    for(auto thr : threads) {
        x += thr->nb_computed.load();
    }
    return x;
}

u32 loct_total() {
    u32 x = 0;
    for(auto thr : threads) {
        x += thr->nb_total.load();
    }
    return x;
}

void loct_calc(std::vector<BeatmapDifficulty*> maps_to_calc) {
    loct_abort();
    if(maps_to_calc.empty()) return;
    if(!cv_normalize_loudness.getBool()) return;

    i32 nb_threads = cv_loudness_calc_threads.getInt();
    if(nb_threads <= 0) {
        nb_threads = env->get_nb_cpu_cores() / 2;
    }
    if(maps_to_calc.size() < nb_threads) nb_threads = maps_to_calc.size();
    int chunk_size = maps_to_calc.size() / nb_threads;
    int remainder = maps_to_calc.size() % nb_threads;

    auto it = maps_to_calc.begin();
    for(int i = 0; i < nb_threads; i++) {
        int cur_chunk_size = chunk_size + (i < remainder ? 1 : 0);

        auto chunk = std::vector<BeatmapDifficulty*>(it, it + cur_chunk_size);
        it += cur_chunk_size;

        auto lct = new LoudnessCalcThread(chunk);
        threads.push_back(lct);
    }
}

void loct_abort() {
    for(auto thr : threads) {
        delete thr;
    }
    threads.clear();
}
